/*		 ICP-PCI Kernel module free public release
 *
 *		       Copyright (C) 1996-99 GOM mbH.
 *	         Rebenring 33, 38106 Braunschweig, GERMANY
 *		            All rights reserved
 *
 * $Id: osLx20.c,v 5.1.2.6 1999/09/22 11:47:12 matthias Exp $
 *
 */

/*
 * Includes:
 */

#include "icpciOS.h"	/* os specific stuff */
#include "icpciExt.h"	/* offical driver interface */
#include "icpciInt.h"	/* driver internals */
#include "icpciReg.h"	/* register definitions IC-PCI */

#include <linux/module.h>	/* init_module */

#ifdef MODULE
/* 
 * Command Line Options (insmod)
 *
 * Bus Master Access:
 * 'bma=disable', or no bma option -> disabled / else enabled
 *
 * Reserved Upper Memory Size
 * 'base=XXM', memory base in MBytes, only if autodetetion off
 * 'size=XXM', memory size in MBytes, only if autodetetion off
 *
 * Byte/Word/Doubleword Register Access Delay Time:
 * 'wait8=20', means 20 usecs or 20 loops (-DICP_LOOP_WAIT) delay
 * 'wait16=5', see above
 * 'wait32=0', see above
 *
 * Major Device Number assignement:
 * 'major0=x',  0       : dynamic assignement (default)
 * 'major1=x',  1 - 128 : explicitely define the number (1 - MAX_CHRDEV)
 *                > 128 : old style behavior (ICP_MAJOR_0|1)
 *
 */

char *bma = "enable";

#ifdef ICP_AUTODETECT
extern unsigned long
bios_memory_end;	/* upper limit of physical memory */

extern unsigned long
memory_end;		/* upper limit of configured memory */
#else
static u_long base = 0;
static size_t size = 0;
#endif

#ifdef ICP_WAIT_LOOP
static int wait8 = 400;
static int wait16 = 200;
static int wait32 = 100;
#else
static int wait8 = 4;
static int wait16 = 2;
static int wait32 = 1;
#endif

static u_int major0 = 0;
static u_int major1 = 0;
#endif

/*
 * Defines
 */

#define ICP_LATENCY_DEF	32	/* default pci bus master latency */

/*
 * Globals:
 */

extern struct 
icp_global_t icp;	/* main driver structure */

char rcsid[] = "version 0.5.1 09/21/1999 (linux 2.0)";

static u_long used_intr = 0UL;	/* mark the used interrupts for the boards */
#define ICP_NUM_X86INTS	16	/* the x86 structure has up to 16 interrupts */

/*
 * Prototypes:
 */

static void
icp_os_intr (int, void*, struct pt_regs*);
static void
icp_os_timeout (u_long);
static void
icp_os_task_queue (void*);
static int
icp_os_open (struct inode*, struct file*);
static void
icp_os_close (struct inode*, struct file*);
static int
icp_os_read (struct inode*, struct file*, char* ,int);
static int
icp_os_write (struct inode*, struct file*, const char* ,int);
static int
icp_os_ioctl (struct inode*, struct file*, u_int, u_long); 
static int
icp_os_mmap (struct inode*, struct file*, struct vm_area_struct*);
static int
icp_os_select (struct inode*, struct file*, int, select_table*);

static struct file_operations icp_fops =
{
  NULL,		/* seek */
  icp_os_read,	/* read */
  icp_os_write,	/* write */
  NULL,		/* readdir */
  icp_os_select,/* select */
  icp_os_ioctl,	/* ioctl */
  icp_os_mmap, 	/* mmap */
  icp_os_open,	/* open */
  icp_os_close,	/* release */
  NULL,		/* fsync */
  NULL,		/* fasync */
  NULL,		/* check_media_change */
  NULL		/* revalidate */
};

/*
 * delayed_memset_io
 */

static void
delayed_memset_io (void *data)
{

  struct icp_memset_t *memset = (struct icp_memset_t*)data;

  while (memset->size > BLOCK_CLEAR_SIZE)
    {
      memset_io (memset->addr, memset->value, BLOCK_CLEAR_SIZE);
      memset->addr += BLOCK_CLEAR_SIZE;
      memset->size -= BLOCK_CLEAR_SIZE;
      if (intr_count)
	/* in irq handler we have to add task queues */
	{
	  icp.osp[memset->board]->clr_t_queue.sync = 0;
	  icp.osp[memset->board]->clr_t_queue.routine = delayed_memset_io;
	  icp.osp[memset->board]->clr_t_queue.data =
	    (void*)&icp.osp[memset->board]->memset;
	  queue_task(&icp.osp[memset->board]->clr_t_queue, &tq_scheduler);
	}
      else
	schedule();
    }

  if (memset->size > 0)
    memset_io (memset->addr, memset->value, memset->size);

  if (intr_count)
    icp_wakeup (memset->board, CAMERA);
}


void*
icp_memset_io (int board, void *addr, char value, size_t size)
{
  
  void *dst_addr = addr;
  
  if (size > BLOCK_CLEAR_SIZE)
    {
      icp.osp[board]->memset.board = board;
      icp.osp[board]->memset.addr = addr;
      icp.osp[board]->memset.value = value;
      icp.osp[board]->memset.size = size;
      delayed_memset_io (&icp.osp[board]->memset);
    }
  else
    if (size > 0)
      memset_io (addr, value, size);
  
  return (dst_addr);
}

/*
 * icp_malloc
 */

void* 
icp_malloc (size_t size, int mode)
{
  switch (mode)
    {
    case UNPAGED:
      return kmalloc (size, GFP_KERNEL);
      break;
    case PAGED:
      return vmalloc (size);
      break;
    default:
      return NULL;
    }
  
  return NULL;
}

/*
 * icp_free
 */

void
icp_free(void *base, int mode)
{
  switch (mode)
    {
    case UNPAGED:
      return kfree (base);
      break;
    case PAGED:
      return vfree (base);
      break;
    }
}

/*
 * icp_add_timer
 */

int
icp_add_timer (int board, time_t timeout, int mode)
{
  struct timer_list *timer_ptr;

  switch (mode)
    {
    case CAMERA:
      timer_ptr = &icp.osp[board]->cam_timer;
      break;
    case IOPORT:
      timer_ptr = &icp.osp[board]->iop_timer;
      break;
    default:
      return -EINVAL;
    }

  if (!del_timer (timer_ptr))
    ICP_PRINTF ("icpci[%c]: timer is not queued.\n", 'W');

  timer_ptr->data = board;
  timer_ptr->function = icp_os_timeout;
  timer_ptr->expires = jiffies + timeout;
  add_timer (timer_ptr);
  
  return OK;
}

/*
 * icp_del_timer
 */

int
icp_del_timer (int board, int mode)
{
  
  struct timer_list *timer_ptr;

  switch (mode)
    {
    case CAMERA:
      timer_ptr = &icp.osp[board]->cam_timer;
      break;
    case IOPORT:
      timer_ptr = &icp.osp[board]->iop_timer;
      break;
    default:
      return -EINVAL;
    }
  
  if (!del_timer (timer_ptr))
    ICP_PRINTF ("icpci[%c]: timer is not queued.\n", 'W');
  
  return OK;
}

/*
 * icp_sleep
 */

int
icp_sleep (int board, time_t timeout, int mode)
{
  
  int error;

  switch (mode)
    {
    case CAMERA:
      if ((error = icp_add_timer (board, timeout, CAMERA)))
	return error;
      interruptible_sleep_on (&icp.osp[board]->cam_w_queue);
      if (current->signal & ~current->blocked)
	return -ERESTARTSYS;
      break;
    case IOPORT:
      if ((error = icp_add_timer (board, timeout, IOPORT)))
	return error;
      interruptible_sleep_on (&icp.osp[board]->iop_w_queue);
      if (current->signal & ~current->blocked)
	return -ERESTARTSYS;
      break;
    default:
      return -EINVAL;
    }

  return OK;
}

/*
 * icp_wakeup
 */

int
icp_wakeup (int board, int mode)
{
  int error;

  switch (mode)
    {
    case CAMERA:
      wake_up (&icp.osp[board]->cam_w_queue);
      if ((error = icp_del_timer (board, CAMERA)))
	return error;
      break;
    case IOPORT:
      wake_up (&icp.osp[board]->iop_w_queue);
      if ((error = icp_del_timer (board, IOPORT)))
	return error;
      break;
    default:
      return -EINVAL;
    }

  return OK;
}

/*
 * icp_marktq
 */

int
icp_marktq (int board)
{

  icp.osp[board]->clr_t_queue.sync = 0;
  icp.osp[board]->clr_t_queue.routine = icp_os_task_queue;
  icp.osp[board]->clr_t_queue.data = (void*)board;
  
  queue_task_irq_off (&icp.osp[board]->clr_t_queue, &tq_scheduler);
  
  return OK;
}

/*
 * icp_irqoff
 */

void
icp_irqoff (int board)
{
  disable_irq (icp.pci[board]->irq_line);
}

/*
 * icp_irqon
 */

void
icp_irqon (int board)
{
  enable_irq (icp.pci[board]->irq_line);
}

/*
 * icp_splhigh
 */

void
icp_splhigh (int board)
{
  save_flags (icp.osp[board]->flags);
  cli();
}

/*
 * icp_spllow
 */

void
icp_spllow (int board)
{
  restore_flags (icp.osp[board]->flags);
}

/*
 * icp_fill_info
 */

int
icp_fill_info (int board, void* dummy_ptr)
{
  struct icp_info_t *frame_info = (struct icp_info_t*)dummy_ptr;
  frame_info->clock_tics = jiffies;
  do_gettimeofday (&(frame_info->time_stamp));
  memcpy (&(frame_info->frame_nums), &(icp.acc[board]),
	  sizeof (struct icp_acc_t));
  
  return OK;
}

/*
 * icp_assign_task
 */

int
icp_assign_task (int board)
{
  icp.osp[board]->task = current;

  return OK;
}

/*
 * icp_check_task
 */

int
icp_task_assigned (int board)
{
  return (icp.osp[board]->task != NULL);
}

/*
 * icp_deassign_task
 */

int
icp_deassign_task (int board)
{
  icp.osp[board]->task = NULL;

  return OK;
}

/*
 * icp_assign_signal
 */

int
icp_assign_signal (int board, icp_flags_t flags, icp_f_owner_t *f_owner)
{
  
#if LINUX_VERSION_CODE < 0x020020
  icp.osp[board]->pid = *f_owner;
#else
  memcpy (&icp.osp[board]->fown, f_owner,
	  sizeof (struct fown_struct));
#endif
  /* enable signal generation if requested */
  if (flags & FASYNC)
    icp.osp[board]->signum = SIGIO;
  else
    icp.osp[board]->signum = 0;
  
  return OK;
}

/*
 * icp_deassign_signal
 */

int
icp_deassign_signal (int board)
{

  icp.osp[board]->signum = 0;
#  if LINUX_VERSION_CODE < 0x020020	/* 2.0.32 */
  icp.osp[board]->pid = 0;
#  else
  icp.osp[board]->fown.pid = 0;
#  endif
  
  return OK;
}

/*
 * icp_check_blocked
 */

int
icp_check_blocked (int board)
{
  return ((icp.osp[board]->task->blocked >> (ICP_SIGIO - 1)) & 1);
}

/*
 * icp_kill_proc
 */

#ifdef DEBUG_ICPCI
#  define DEBUG_ICPCI_IRQ
#endif

int
icp_kill_proc (int board)
{

  if (icp.osp[board]->signum == ICP_SIGIO &&
#if LINUX_VERSION_CODE < 0x020020	/* 2.0.32 */
      icp.osp[board]->pid != 0)
#else
      icp.osp[board]->fown.pid != 0)
#endif
    {
#ifdef DEBUG_ICPCI_IRQ
      icp_irq_printf ("<s>");
#endif
#if LINUX_VERSION_CODE < 0x020020
      kill_proc (icp.osp[board]->pid, ICP_SIGIO, 1);
#else
      kill_proc (icp.osp[board]->fown.pid, ICP_SIGIO, 1);
#endif
    }
  
  return OK;
}

/*
 * icp_os_intr
 */

void
icp_os_intr (int irq, void *dev_id, struct pt_regs *regs)
{
  icp_intr (irq);
}

/*
 * icp_os_timeout
 */

void
icp_os_timeout (ulong data)
{
  icp_timeout ((int)data);
}

/*
 * icp_os_bottomhalf
 */

void
icp_os_task_queue (void *data)
{
  icp_task_queue ((int)data);
}

/*
 * icp_os_open
 */

static int
icp_os_open (struct inode *inode, struct file *file)
{
  
  int error;
  
  struct icp_devno_t devno;
  
  devno.major = MAJOR(inode->i_rdev);
  devno.major0 = major0;	/* dynamic (or explicit) assigned */
  devno.major1 = major1;	/* dynamic (or explicit) assigned */
  devno.minor = MINOR(inode->i_rdev);
  
  if ((error = icp_decode_devno (&devno)))
    return error;
  
  if ((error = icp_open (&devno, file->f_flags)))
    return error;
  
#ifdef MODULE    
  MOD_INC_USE_COUNT;
#endif    
  
  return OK;
}

/*
 * icp_os_close
 */

static void
icp_os_close (struct inode *inode, struct file *file)
{
  
  struct icp_devno_t devno;

  devno.major = MAJOR(inode->i_rdev);
  devno.major0 = major0;	/* dynamic (or explicit) assigned */
  devno.major1 = major1;	/* dynamic (or explicit) assigned */
  devno.minor = MINOR(inode->i_rdev);
  
  if (icp_decode_devno (&devno))
    return;
  
  icp_close (&devno, file->f_flags);
  
#ifdef MODULE    
  MOD_DEC_USE_COUNT;
#endif    
}

/*
 * icp_os_read
 */

static int
icp_os_read (struct inode *inode, struct file *file,
	     char *buffer, int count)
{

  int error;
  
  struct icp_devno_t devno;
  
  if (verify_area(VERIFY_WRITE, buffer, count))
    return -EFAULT;
  
  devno.major = MAJOR(inode->i_rdev);
  devno.major0 = major0;	/* dynamic (or explicit) assigned */
  devno.major1 = major1;	/* dynamic (or explicit) assigned */
  devno.minor = MINOR(inode->i_rdev);

  if ((error = icp_decode_devno (&devno)))
    return error;
  
  return icp_read (&devno, file->f_flags, buffer, count);
}

/*
 * icp_os_write
 */

static int
icp_os_write (struct inode *inode, struct file *file,
              const char *buffer, int count)
{

  int error;
  
  struct icp_devno_t devno;
  
  if (verify_area(VERIFY_WRITE, buffer, count))
    return -EFAULT;
  
  devno.major = MAJOR(inode->i_rdev);
  devno.major0 = major0;	/* dynamic (or explicit) assigned */
  devno.major1 = major1;	/* dynamic (or explicit) assigned */
  devno.minor = MINOR(inode->i_rdev);

  if ((error = icp_decode_devno (&devno)))
    return error;
  
  return icp_write (&devno, file->f_flags, buffer, count);
}

/*
 * icp_os_mmap
 */

static int
icp_os_mmap (struct inode *inode, struct file *file,
	     struct vm_area_struct *vma)
{
  
  int error;
  
  struct icp_devno_t devno;

  u_long base;  
  size_t size;
  
  devno.major = MAJOR(inode->i_rdev);
  devno.major0 = major0;	/* dynamic (or explicit) assigned */
  devno.major1 = major1;	/* dynamic (or explicit) assigned */
  devno.minor = MINOR(inode->i_rdev);
  
  if ((error = icp_decode_devno (&devno)))
    return error;
  
  size = vma->vm_end - vma->vm_start;
  
  base = 0UL;
  if ((error = icp_mmap (&devno, &base, size)))
    return error;
  if (base == 0UL)
    return -EFAULT;

  /* remapping */
  if ((error = remap_page_range(vma->vm_start, base,
				size, vma->vm_page_prot)))
    return error;
  
  vma->vm_inode = inode;
  inode->i_count++;

  return OK;
}

/*
 * icp_os_select
 */

int
icp_os_select (struct inode *inode, struct file *file,
	       int type, select_table *wait)
{
  
  int error, result;
  
  struct icp_devno_t devno;

  devno.major = MAJOR(inode->i_rdev);
  devno.major0 = major0;	/* dynamic (or explicit) assigned */
  devno.major1 = major1;	/* dynamic (or explicit) assigned */
  devno.minor = MINOR(inode->i_rdev);
  
  if ((error = icp_decode_devno (&devno)))
    return error;

  ICP_PRINTF("icpci[%c]: icp_select - %s.\n", 'F', "TEST  IN|EX");
  ICP_PRINTF("icpci[%d]:       time : %ld.\n", devno.board, jiffies);
    
  result = icp_select (&devno, type);
  
  if (type & ICP_SEL_IN)
    {
      if (!result)
	{
	  ICP_PRINTF("icpci[%c]: icp_select - %s.\n", 'F', "WAIT  IN");
	  select_wait (&icp.osp[devno.board]->cam_w_queue, wait);
	}
      else
	if (result & ICP_SEL_IN)
	  {
	    ICP_PRINTF("icpci[%c]: icp_select - %s.\n", 'F', "FOUND IN");
	    return TRUE;
	  }
    }
  else
    if ((type & SEL_EX) && (result & ICP_SEL_EX))
      {
	ICP_PRINTF("icpci[%c]: icp_select - %s.\n", 'F', "FOUND EX");
	return TRUE;
      }

  return FALSE;
}

/* 
 * icp_os_ioctl
 */

int
icp_os_ioctl (struct inode *inode, struct file *file,
	      u_int cmd, u_long arg)
{

  int error, count;
  
  struct icp_devno_t devno;

  union icp_inparam_t *inparam = NULL;
  union icp_outparam_t *outparam = NULL;
  
  devno.major = MAJOR(inode->i_rdev);
  devno.major0 = major0;	/* dynamic (or explicit) assigned */
  devno.major1 = major1;	/* dynamic (or explicit) assigned */
  devno.minor = MINOR(inode->i_rdev);

  if ((error = icp_decode_devno (&devno)))
    return error;
  
  /* validate module specific commands */
  if (_IOC_NR(cmd) > 31 && _IOC_NR(cmd) < 224)
    switch (icp.cap[devno.board]->module)
      {
      case ICP_AMFA:
	if (_IOC_NR(cmd) > 63 || _IOC_TYPE(cmd) != 'G')
	  return -EINVAL;
	break;
      case ICP_AMC1:
	if (_IOC_NR(cmd) < 64 || _IOC_NR(cmd) > 95 || _IOC_TYPE(cmd) != 'G')
	  return -EINVAL;
      break;
      case ICP_AMVS:
	if (_IOC_NR(cmd) < 96 || _IOC_NR(cmd) > 127 || _IOC_TYPE(cmd) != 'G')
	  return -EINVAL;
	break;
      case ICP_AMDG:
	if (_IOC_NR(cmd) < 128 || _IOC_NR(cmd) > 159 || _IOC_TYPE(cmd) != 'G')
	  return -EINVAL;
	break;
      case ICP_AMPV:
	if (_IOC_NR(cmd) < 96 || _IOC_NR(cmd) > 127 || _IOC_TYPE(cmd) != 'H')
	  return -EINVAL;
	break;
      default:
	return -ENXIO;
      }

  /* structure allocation */
  if (_IOC_DIR(cmd) == _IOC_READ ||
      _IOC_DIR(cmd) == (_IOC_READ | _IOC_WRITE))
    outparam = icp_malloc (sizeof(union icp_outparam_t), PAGED);

  if (_IOC_DIR(cmd) == _IOC_WRITE ||
      _IOC_DIR(cmd) == (_IOC_WRITE | _IOC_READ))
    inparam = icp_malloc (sizeof(union icp_inparam_t), PAGED);

  if (_IOC_DIR(cmd) == _IOC_READ)
    {
      if (cmd == GIOC_DG_GET_LUT)
	/* data has to be copied in single-page chunks */
	{
	  if (verify_area (VERIFY_WRITE, (void *)arg,
			   sizeof(union dg_lut16_t)))
	    return -EFAULT;
	}
      else
	{
	  if (verify_area (VERIFY_WRITE,(void *)arg, _IOC_SIZE(cmd)))
	    return -EFAULT;
	}
    }

  /* copyin */
  if (_IOC_DIR(cmd) == _IOC_WRITE ||
      _IOC_DIR(cmd) == (_IOC_WRITE | _IOC_READ))
    {
      if (cmd == GIOC_DG_SET_LUT)
	{
	  /* data has to be copied in single-page chunks */
	  for (count = 0; count < (sizeof(union dg_lut16_t) / PAGE_SIZE);
	       count++)
	    icp_copyin ((void *)inparam + count * PAGE_SIZE, 
			(void *)arg + count * PAGE_SIZE, PAGE_SIZE);
	  /* if the structure is not a multiple of PAGE_SIZE */
	  if (sizeof(union dg_lut16_t) % PAGE_SIZE)
	    icp_copyin ((void *)inparam + count * PAGE_SIZE, 
			(void *)arg + count * PAGE_SIZE,
			sizeof(union dg_lut16_t) % PAGE_SIZE);
      }
      else
	{
	icp_copyin (inparam, (void *)arg, _IOC_SIZE(cmd));
	}
    }
  
  if ((error = icp_ioctl (&devno, file->f_flags, &file->f_owner,
			  cmd, inparam, outparam)))
    {
      if (outparam != NULL)
	icp_free (outparam, PAGED);
      if (inparam != NULL)
	icp_free (inparam, PAGED);
      return error;
    }

  /* copyout */
  if (_IOC_DIR(cmd) == _IOC_READ ||
      _IOC_DIR(cmd) == (_IOC_WRITE | _IOC_READ))
    {
      if (cmd == GIOC_DG_GET_LUT)
	{
	  /* data has to be copied in single-page chunks */
	  for (count = 0; count < (sizeof(union dg_lut16_t) / PAGE_SIZE);
	       count++)
	    icp_copyout ((void *)outparam + count * PAGE_SIZE, 
			 (void *)arg + count * PAGE_SIZE, PAGE_SIZE);
	  /* if the structure is not a multiple of PAGE_SIZE */
	  if (sizeof(union dg_lut16_t) % PAGE_SIZE)
	    icp_copyout ((void *)outparam + count * PAGE_SIZE, 
			 (void *)arg + count * PAGE_SIZE,
			 sizeof(union dg_lut16_t) % PAGE_SIZE);
	}
      else
	{
	  icp_copyout ((void *)arg, outparam, _IOC_SIZE(cmd));
	}
    }

  /* free structures */
  if (outparam != NULL)
    icp_free (outparam, PAGED);
  if (inparam != NULL)
    icp_free (inparam, PAGED);

  return OK;
}

int
icp_os_init (int board)
{

  ICP_PRINTF("icpci[%c]: icp_os_init.\n", 'F');

  /* init os layer */
  icp.osp[board] =
    (struct icp_osp_t*)icp_malloc (sizeof (struct icp_osp_t), UNPAGED);
  if (icp.osp[board] == NULL)
    return -ENOMEM;
  memset (icp.osp[board], 0x00, sizeof(struct icp_osp_t));
  
  /* timer init */
  init_timer (&icp.osp[board]->cam_timer);
  init_timer (&icp.osp[board]->iop_timer);

  return OK;
}

/*
 * icp_os_pci_init
 */

int
icp_os_pci_init (int *boards) {

  int	error;
  int	board;

  u_char	icp_bus,
    		icp_dev_fn,
    		irq_line,
    		lat_tim,
		rev_id,
    		lev_id;

  u_short	icp_index,
    		icp_command;

  u_long	p_icr_base,
    		p_cr_base,
    		p_mod_base,
    		p_mem_base;

  ICP_PRINTF("icpci[%c]: icp_os_pci_init.\n", 'F');

  /* bios installed ? */

  if (!pcibios_present()) {
    icp_printf ("icpci[%c]: no 'pci bios' present.\n", 'P');
    return (-ENXIO);
  }

  for (board = 0; board < ICP_BOARDS_MAX; board++) {
    icp_index = board;
    /* find board and allocate structure */
    if (pcibios_find_device (PCI_VENDOR_ID_IMAGINGTECH,
			     PCI_DEVICE_ID_IMAGINGTECH_ICPCI,
			     icp_index, &icp_bus, &icp_dev_fn)
	!= PCIBIOS_SUCCESSFUL)
      break;
    else {
      icp.pci[board] =
	(struct icp_pci_t*)icp_malloc (sizeof (struct icp_pci_t), UNPAGED);
      if (icp.pci[board] == NULL) {
	return (-ENOMEM);
      }
      memset (icp.pci[board], 0x00, sizeof(struct icp_pci_t));
    }

    /* read revision id */
    error = pcibios_read_config_byte (icp_bus, icp_dev_fn,
				      PCI_REVISION_ID,
				      &rev_id);
    if (error != PCIBIOS_SUCCESSFUL) {
      icp_printf ("icpci[%d]: cannot read revision id.\n", board);
      return (-ENXIO);
    }

    /* read revision level id from status register (??)*/
    error = pcibios_read_config_byte (icp_bus, icp_dev_fn,
				      PCI_STATUS,
				      &lev_id);
    if (error != PCIBIOS_SUCCESSFUL) {
      icp_printf ("icpci[%d]: cannot read status register.\n", board);
      return (-ENXIO);
    }
    
    /* base address of interface control register set */
    error = pcibios_read_config_dword (icp_bus, icp_dev_fn,
				       PCI_BASE_ADDRESS_0,
				       (unsigned int*)&p_icr_base);
    if (error != PCIBIOS_SUCCESSFUL) {
      icp_printf ("icpci[%d]: cannot read base address 0.\n", board);
      return (-EIO);
    }

    /* base address of control register set */
    error = pcibios_read_config_dword (icp_bus, icp_dev_fn,
				       PCI_BASE_ADDRESS_1,
				       (unsigned int*)&p_cr_base);
    if (error != PCIBIOS_SUCCESSFUL) {
      icp_printf ("icpci[%d]: cannot read base address 1.\n", board);
      return (-ENXIO);
    }

    /* base address of acquisition module register set */
    error = pcibios_read_config_dword (icp_bus, icp_dev_fn,
				       PCI_BASE_ADDRESS_2,
				       (unsigned int*)&p_mod_base);
    if (error != PCIBIOS_SUCCESSFUL) {
      icp_printf ("icpci[%d]: cannot read base address 2.\n", board);
      return (-ENXIO);
    }

    /* base address of frame grabber on board memory */
    error = pcibios_read_config_dword (icp_bus, icp_dev_fn,
				       PCI_BASE_ADDRESS_3,
				       (unsigned int*)&p_mem_base);
    if (error != PCIBIOS_SUCCESSFUL) {
      icp_printf ("icpci[%d]: cannot read base address 3.\n", board);
      return (-ENXIO);
    }

    /* interrupt line */
    error = pcibios_read_config_byte (icp_bus, icp_dev_fn,
				      PCI_INTERRUPT_LINE,
				      &irq_line);
    if (error != PCIBIOS_SUCCESSFUL) {
      icp_printf ("icpci[%d]: cannot read interrupt line.\n", board);
      return (-ENXIO);
    }
    if (irq_line < 3 || irq_line > 15 )
      icp_printf ("icpci[%d]: really use interrupt %d?\n", board, irq_line);
    
    /* latency timer (if too low -> correct) */
    error = pcibios_read_config_byte (icp_bus, icp_dev_fn,
				      PCI_LATENCY_TIMER,
				      &lat_tim);
    if (error != PCIBIOS_SUCCESSFUL) {
      icp_printf ("icpci[%d]: cannot read latency timer.\n", board);
      return (-ENXIO);
    }
    if (lat_tim < ICP_LATENCY_DEF) {
      error = pcibios_write_config_byte (icp_bus, icp_dev_fn,
					 PCI_LATENCY_TIMER,
					 ICP_LATENCY_DEF);
      if (error != PCIBIOS_SUCCESSFUL) {
	icp_printf ("icpci[%d]: cannot write latency timer.\n", board);
	return (-ENXIO);
      }
    }

    /* command register (memory access, i/o access, bus master mode) */
    error = pcibios_read_config_word (icp_bus, icp_dev_fn,
				      PCI_COMMAND,
				      &icp_command);
    if (error != PCIBIOS_SUCCESSFUL) {
      icp_printf ("icpci[%d]: cannot read command register.\n", board);
      return (-ENXIO);
    }
    icp_command |= (PCI_COMMAND_IO |
		    PCI_COMMAND_MEMORY |
		    PCI_COMMAND_MASTER);
    error = pcibios_write_config_word (icp_bus, icp_dev_fn,
				       PCI_COMMAND,
				       icp_command);
    if (error != PCIBIOS_SUCCESSFUL) {
      icp_printf ("icpci[%d]: cannot write command register.\n", board);
      return (-ENXIO);
    }

    /* debugging */
    ICP_PRINTF("icpci[%d]: Bus       : %d.\n", board, icp_bus);
    ICP_PRINTF("icpci[%d]: Device    : %d.\n", board, PCI_SLOT(icp_dev_fn));
    ICP_PRINTF("icpci[%d]: Function  : %d.\n", board, PCI_FUNC(icp_dev_fn));
    ICP_PRINTF("icpci[%d]: Index     : %d.\n", board, icp_index);
    ICP_PRINTF("icpci[%d]: Latency   : %d.\n", board, lat_tim);

    /* fill in structure */
    icp.pci[board]->address0 = p_icr_base;
    icp.pci[board]->address1 = p_cr_base;
    icp.pci[board]->address2 = p_mod_base;
    icp.pci[board]->address3 = p_mem_base;
    icp.pci[board]->irq_line = irq_line;
    icp.pci[board]->revision = rev_id;
    icp.pci[board]->status = lev_id;
    
    (*boards)++;
  }

  /* FIXME: for linux we use reserved upper memory */
  /* for the bigphysarea-patch we have to modify the kernel */
  /* the best way is to do a vmalloc_continuous (not available) */

  /* remap the reserved upper memory space (if available) */
  if (icp.bma_on && *boards)
    {
      /* get buffer begin and size in two different ways */
#ifdef ICP_AUTODETECT
      if ((icp.bma_size = bios_memory_end - memory_end) == 0UL)
	{
	  icp_printf ("icpci[%c]: no unused upper memory, "
		      "bma disabled!\n", 'W');
	  icp.bma_on = FALSE;
	  return OK;
	}
      else
	icp.p_bma_base = memory_end;
#else
      if (size)
	icp.bma_size = size * 1024 * 1024;
      else
	{
	  icp_printf ("icpci[%c]: no unused upper memory, "
		      "bma disabled!\n", 'W');
	  icp.bma_on = FALSE;
	  return OK;
	}
      if (base > 4)
	icp.p_bma_base = base * 1024 * 1024;
      else
	{
	  icp_printf ("icpci[%c]: no or wrong base address, "
		      "bma disabled!\n", 'W');
	  icp.bma_on = FALSE;
	  return OK;
	}
#endif
  
      /* begin and size are always aligned (multilple of 1Mb) */
      if ((icp.v_bma_base = vremap (icp.p_bma_base, icp.bma_size)) == NULL)
	{
	  icp_printf ("icpci[%d]: unable to vremap unused upper memory.\n",
		      board);
	  return -EFAULT;
	}
    }

  return OK;
}

/*
 * icp_os_map_init
 */

int
icp_os_map_init (int board, u_long *align_div, ptrdiff_t *align_mod)
{

  void	 	*v_icr_base = NULL;
  void		*v_cr_base = NULL;
  void		*v_mod_base = NULL;
  void		*v_mem_base = NULL;

  ICP_PRINTF("icpci[%c]: icp_os_map_init.\n", 'F');

  /* remap all register spaces and on board memory to low kernel memory */
  if (!icp.map[board]->regs_io) {
    if ((v_icr_base = vremap (align_div[ICR_REGS],
			      ROUND_UPTO_PAGE(align_mod[ICR_REGS] + 
					      ICR_REG_SIZE *
					      sizeof(u_long)))) == NULL) {
      icp_printf ("icpci[%d]: unable to ioremap interface control regs.\n",
		  board);
      return (-EFAULT);
    }
    if ((v_cr_base = vremap (align_div[CR_REGS],
			     ROUND_UPTO_PAGE(align_mod[CR_REGS] + 
					     CR_REG_SIZE *
					     sizeof (u_short)))) == NULL) {
      icp_printf ("icpci[%d]: unable to ioremap control regs.\n", board);
      return (-EFAULT);
    }
    icp.map[board]->v_icr_base = v_icr_base;
    icp.map[board]->v_cr_base = v_cr_base;
  }
  if ((v_mod_base = vremap (icp.pci[board]->address2,
			    MOD_MAP_SIZE)) == NULL) {
    icp_printf ("icpci[%d]: unable to ioremap acquisition modul regs.\n",
		board);
    return (-EFAULT);
  }
  if (icp.map[board]->big_mem) {
    if ((v_mem_base = vremap (icp.pci[board]->address3,
			     MEM_MAP_SIZE << 1)) == NULL) {
      icp_printf ("icpci[%d]: unable to ioremap main board memory.\n",
		  board);
      return (-EFAULT);
    }
  } else {
    if ((v_mem_base = vremap (icp.pci[board]->address3,
			     MEM_MAP_SIZE)) == NULL) {
      icp_printf ("icpci[%d]: unable to ioremap main board memory.\n",
		  board);
      return (-EFAULT);
    }
  }
  icp.map[board]->v_mod_base = v_mod_base;
  icp.map[board]->v_mem_base = v_mem_base;

  return (OK);
}

/*
 * icp_os_cap_init
 */

int
icp_os_cap_init (int board)
{

  icp.cap[board]->io_wait8 = icp.cap[board]->mem_wait8 = wait8;
  icp.cap[board]->io_wait16 = icp.cap[board]->mem_wait16 = wait16;
  icp.cap[board]->io_wait32 = icp.cap[board]->mem_wait32 = wait32;

  return OK;
}

#ifdef MODULE
/*
 * cleanup_module
 */

void
cleanup_module (void)
{  
  
  int	intr, board, camera;
  
  ICP_PRINTF("icpci[%c]: cleanup_module.\n", 'F');
  
  for (board = 0; board < ICP_BOARDS_MAX; board++ )
    {
      if (icp.bma[board] != NULL)
	{
	  ICP_PRINTF("icp.bma[%d]: 0x%p\n", board, icp.bma[board]);
	  icp_free (icp.bma[board], UNPAGED);
	}
      if (icp.mem[board] != NULL)
	{
	  ICP_PRINTF("icp.mem[%d]: 0x%p\n", board, icp.mem[board]);
	  icp_free (icp.mem[board], UNPAGED);
	}
      for (camera = 0; camera < ICP_CAMERAS_MAX; camera++)
	if (icp.dev[board][camera] != NULL)
	  {
	    ICP_PRINTF("icp.dev[%d][%d]: 0x%p\n", board, camera,
		   icp.dev[board][camera]);
	    icp_free (icp.dev[board][camera], UNPAGED);
	  }
      if (icp.cap[board] != NULL)
	{
	  ICP_PRINTF("icp.cap[%d]: 0x%p\n", board, icp.cap[board]);
	  icp_free (icp.cap[board], UNPAGED);
	}
      if (icp.map[board] != NULL)
	{
	  if (icp.map[board]->v_mem_base != NULL)
	    {
	      ICP_PRINTF("icp.map[%d]->v_mem_base: 0x%p\n", board,
		     icp.map[board]->v_mem_base);
	      vfree (icp.map[board]->v_mem_base);
	    }
	  if (icp.map[board]->v_mod_base != NULL)
	    {
	      ICP_PRINTF("icp.map[%d]->v_mod_base: 0x%p\n", board,
		     icp.map[board]->v_mod_base);
	      vfree (icp.map[board]->v_mod_base);
	    }
	  if (icp.map[board]->v_cr_base != NULL)
	    {
	      icp.map[board]->v_cr_base =
		(u_short*)ROUND_DOWNTO_PAGE(icp.map[board]->v_cr_base);
	      ICP_PRINTF("icp.map[%d]->v_cr_base: 0x%p\n", board,
		     icp.map[board]->v_cr_base);
	      vfree (icp.map[board]->v_cr_base);
	    }
	  if (icp.map[board]->v_icr_base != NULL)
	    {
	      icp.map[board]->v_icr_base =
		(u_long*)ROUND_DOWNTO_PAGE(icp.map[board]->v_icr_base);
	      ICP_PRINTF("icp.map[%d]->v_icr_base: 0x%p\n", board,
		     icp.map[board]->v_icr_base);
	      vfree (icp.map[board]->v_icr_base);
	    }
	  ICP_PRINTF("icp.map[%d]: 0x%p\n", board, icp.map[board]);
	  icp_free (icp.map[board], UNPAGED);
	}
      if (icp.osp[board] != NULL)
	{
	  ICP_PRINTF("icp.osp[%d]: 0x%p\n", board,
		     icp.osp[board]);
	  icp_free (icp.osp[board], UNPAGED);
	}
      if (icp.pci[board] != NULL) {
	ICP_PRINTF("icp.pci[%d]: 0x%p\n", board, icp.pci[board]);
	icp_free (icp.pci[board], UNPAGED);
      }
    }
  if (icp.v_bma_base != NULL)
    {
      ICP_PRINTF("icp.v_bma_base[%d]: 0x%p\n", board, icp.v_bma_base);
      vfree (icp.v_bma_base);
    }

  for (intr = 0; intr < ICP_NUM_X86INTS; intr++)
    if (used_intr & (1 << intr))
      {
	free_irq (intr, &icp);
	used_intr &= ~(1 << intr);
      }

  if (unregister_chrdev(major1, "icpciB") < 0)
    icp_printf ("icpci[%c]: error while unregistering 'icpciB' (%d).\n",
		'E', major1);
  if (unregister_chrdev(major0, "icpciA") < 0)
    icp_printf ("icpci[%c]: error while unregistering 'icpciA' (%d).\n",
		'E', major0);
}

/*
 * init_module
 */

int
init_module (void)
{
  
  int	error, board, boards;
  
  ICP_PRINTF("icpci[%c]: init_module.\n", 'F');
  
  /* how are the major devno's assigned (default is 0 - dynamic) */
  if (major0 != 0)
    {
      /* non automatic */
      if (major0 > MAX_CHRDEV)
	{
	  icp_printf ("icpci[%c]: major device number %d out of range "
		      "(1-%d).\n",
		      'I', major0, MAX_CHRDEV);
	  icp_printf ("icpci[%c]: using device number %d instead.\n",
		      'I', ICP_MAJOR_0);
	  major0 = ICP_MAJOR_0;	/* old driver behavior */
	}
      /* register first character device */
      if ((error = register_chrdev (major0, "icpciA", &icp_fops)) < 0)
	{
	  icp_printf ("icpci[%c]: 'register_chrdev (character device %d)' "
		      "failed.\n", 'I', major0);
	  cleanup_module();
	  return error;
	}
    }
  else
    {
      /* register first character device */
      if ((major0 = register_chrdev (major0, "icpciA", &icp_fops)) < 0)
	{
	  icp_printf ("icpci[%c]: 'register_chrdev (character device %d)' "
		      "failed.\n", 'I', major0);
	  cleanup_module();
	  return major0;
	}
    }  
  if (major1 != 0)
    {
      if (major1 > MAX_CHRDEV)
	{
	  /* non automatic */
	  icp_printf ("icpci[%c]: major device number %d out of range "
		      "(1-%d).\n",
		      'I', major1, MAX_CHRDEV);
	  icp_printf ("icpci[%c]: using device number %d instead.\n",
		      'I', ICP_MAJOR_1);
	  major1 = ICP_MAJOR_1;	/* old driver behavior */
	}
      /* register second character device */
      if ((error = register_chrdev (major1, "icpciB", &icp_fops)) < 0)
	{
	  icp_printf ("icpci[%c]: 'register_chrdev (character device %d)' "
		      "failed.\n", 'I', major1);
	  cleanup_module();
	  return error;
	}
    }
  else
    {
      /* register second character device */
      if ((major1 = register_chrdev (major1, "icpciB", &icp_fops)) < 0)
	{
	  icp_printf ("icpci[%c]: 'register_chrdev (character device %d)' "
		      "failed.\n", 'I', major1);
	  cleanup_module();
	  return major1;
	}
    }

  /* parse command line options */
  if (!strcmp (bma, "enable"))
    {
      icp.bma_on = TRUE;
      icp_printf ("icpci[%c]: bus master mode.\n", 'I');
    }
  else
    {
      icp.bma_on = FALSE;
      icp_printf ("icpci[%c]: normal memcpy mode.\n", 'I');	
    }

  /* check PCI - configuration space for possible devices */
  boards = 0;
  if ((error = icp_pci_init (&boards)) < 0)
    {
      cleanup_module();
      return error;
    }
  else
    if (boards == 0)
      {
	icp_printf ("icpci[%c]: no devices found.\n", 'I');
	{
	  cleanup_module();
	  return -ENXIO;
	}
      }
    else
      {
	icp_printf ("icpci[%c]: %d device(s) found.\n", 'I', boards);
	icp.boards = boards;
      }

  /* initialize data structures */
  for (board = 0; board < boards; board++)
    {    
      /* init at first the os structure */
      if ((error = icp_os_init (board)))
	{
	  cleanup_module();
	  return error;
	}
      if ((error = icp_map_init (board)))
	{
	  cleanup_module();
	  return error;
	}
      if ((error = icp_cap_init (board)))
	{
	  cleanup_module();
	  return error;
	}
      if ((error = icp_dev_init (board)))
	{
	  cleanup_module();
	  return error;
	}
    }
  
  for (board = 0; board < boards; board++)
    {
      if ((error = icp_mem_init (board)))
	{
	  cleanup_module();
	  return error;
	}
      if (icp.bma_on)
	if ((error = icp_bma_init (board)))
	  {
	    cleanup_module();
	    return error;
	  }
      /* reset statistics */
      icp.acc[board].frames_captured = 0UL;
      icp.acc[board].frames_valid = 0UL;
      icp.acc[board].frames_lost = 0UL;
      icp.acc[board].events_captured = 0UL;
      icp.acc[board].bytes_valid = 0UL;
      icp.acc[board].bytes_lost = 0UL;
      /* init on-board io-ports (newer boards only) */
      if (icp.pci[board]->revision > ICP_REV_A)
	if ((error = icp_iop_init (board)))
	  {
	    cleanup_module();
	    return error;
	  }
    }

  /* register per board interrupts */
  for (board = 0; board < boards; board++)
    if (!(used_intr & (1 << icp.pci[board]->irq_line)))
      {
	if ((error = request_irq (icp.pci[board]->irq_line,
				  &icp_os_intr,
				  SA_SHIRQ,
				  "icpci", &icp)))
	  if ((error = request_irq (icp.pci[board]->irq_line,
				    &icp_os_intr,
				    SA_INTERRUPT | SA_SHIRQ,
				    "icpci", &icp)))
	    {
	      icp_printf ("icpci[%c]: 'request_irq (intr. %d)' failed.\n",
			  'I', icp.pci[board]->irq_line);
	      cleanup_module();
	      return error;
	    }
	used_intr |= (1 << icp.pci[board]->irq_line);
      }
  
  /* print version */
  icp_printf ("icpci[%c]: %s\n", 'I', rcsid);

  return OK;
}
#endif /* MODULE */

